"""State and remote services of one vehicle."""

import datetime
import logging
from typing import Optional

from pysmarthashtag.models import ValueWithUnit, get_element_from_dict_maybe
from pysmarthashtag.vehicle.battery import Battery
from pysmarthashtag.vehicle.climate import Climate
from pysmarthashtag.vehicle.maintenance import Maintenance
from pysmarthashtag.vehicle.position import Position
from pysmarthashtag.vehicle.running import Running
from pysmarthashtag.vehicle.safety import Safety
from pysmarthashtag.vehicle.tires import Tires

_LOGGER = logging.getLogger(__name__)


class SmartVehicle:
    """Models state and remote services of one vehicle.

    :param account: The account associated with the vehicle.
    :param attributes: attributes of the vehicle as provided by the server.
    """

    data: dict
    """The raw data of the vehicle."""

    odometer: Optional[ValueWithUnit] = None
    """The odometer of the vehicle."""

    battery: Optional[Battery] = None
    """The battery of the vehicle."""

    tires: Optional[Tires] = None
    """The tires of the vehicle."""

    position: Optional[Position] = None
    """The position of the vehicle."""

    last_update: Optional[datetime.datetime] = None
    """The last time the vehicle data was updated."""

    service: Optional[dict] = {}

    maintenance: Optional[Maintenance] = None
    """The maintenance status of the vehicle."""

    running: Optional[Running] = None
    """The running status of the vehicle."""

    climate: Optional[Climate] = None
    """The climate status of the vehicle."""

    safety: Optional[Safety] = None
    """The safety status of the vehicle."""

    climate_control: Optional["ClimateControll"] = None  # noqa: F821

    engine_state: Optional[str] = None
    """The state of the engine."""

    def __init__(
        self,
        account: "SmartAccount",  # noqa: F821
        vehicle_base: dict,
        vehicle_state: Optional[dict] = None,
        charging_settings: Optional[dict] = None,
        fetched_at: Optional[datetime.datetime] = None,
    ) -> None:
        """Initialize the vehicle."""
        self.account = account
        self.data = {}
        self.combine_data(vehicle_base, vehicle_state, charging_settings, fetched_at)
        _LOGGER.debug(
            "Initialized vehicle %s (%s)",
            self.name,
            self.vin,
        )

    def combine_data(
        self,
        vehicle_base: dict,
        vehicle_state: Optional[dict] = None,
        charging_settings: Optional[dict] = None,
        fetched_at: Optional[datetime.datetime] = None,
    ) -> dict:
        """Combine all data into one dictionary."""
        self.data.update(vehicle_base)
        if vehicle_state:
            self.data.update(vehicle_state)
        if charging_settings:
            self.data.update(charging_settings)
        if fetched_at:
            self.data["fetched_at"] = fetched_at
        self._parse_data()
        self.battery = Battery.from_vehicle_data(self.data)
        self.tires = Tires.from_vehicle_data(self.data)
        self.position = Position.from_vehicle_data(self.data)
        self.maintenance = Maintenance.from_vehicle_data(self.data)
        self.running = Running.from_vehicle_data(self.data)
        self.climate = Climate.from_vehicle_data(self.data)
        self.safety = Safety.from_vehicle_data(self.data)

        from pysmarthashtag.control.climate import ClimateControll

        self.climate_control = ClimateControll(self.account.config, self.vin)

    def _parse_data(self) -> None:
        self.vin = self.data.get("vin")
        self.name = self.data.get("modelName")
        odometer = get_element_from_dict_maybe(
            self.data, "vehicleStatus", "additionalVehicleStatus", "maintenanceStatus", "odometer"
        )
        if odometer:
            self.odometer = ValueWithUnit(
                int(float(odometer)),
                "km",
            )
        last_update = get_element_from_dict_maybe(self.data, "vehicleStatus", "updateTime")
        if last_update:
            self.last_update = datetime.datetime.fromtimestamp(int(last_update) / 1000, datetime.timezone.utc)
        days_to_service = get_element_from_dict_maybe(
            self.data, "vehicleStatus", "additionalVehicleStatus", "maintenanceStatus", "daysToService"
        )
        distance_to_service = get_element_from_dict_maybe(
            self.data, "vehicleStatus", "additionalVehicleStatus", "maintenanceStatus", "distanceToService"
        )
        self.service["daysToService"] = int(days_to_service) if days_to_service else None
        self.service["distanceToService"] = ValueWithUnit(distance_to_service, "km") if distance_to_service else None

        self.engine_state = get_element_from_dict_maybe(
            self.data, "vehicleStatus", "basicVehicleStatus", "engineStatus"
        )



"""Access to Smart account for your vehicles therin."""

import datetime
import json
import logging
from dataclasses import InitVar, dataclass, field
from typing import Dict

from pysmarthashtag.api import utils
from pysmarthashtag.api.authentication import SmartAuthentication
from pysmarthashtag.api.client import SmartClient, SmartClientConfiguration
from pysmarthashtag.const import API_BASE_URL, API_CARS_URL, API_SELECT_CAR_URL
from pysmarthashtag.models import SmartAuthError, SmartHumanCarConnectionError, SmartTokenRefreshNecessary
from pysmarthashtag.vehicle.vehicle import SmartVehicle

VALID_UNTIL_OFFSET = datetime.timedelta(seconds=10)

_LOGGER = logging.getLogger(__name__)


@dataclass
class SmartAccount:
    """Create a new connection to the Smart web service."""

    username: str
    """Username for the Smart account."""

    password: InitVar[str]
    """Password for the Smart account."""

    config: SmartClientConfiguration = None
    """Configuration for the Smart client."""

    log_responses: InitVar[bool] = False
    """Optional. If set, all responses from the server will be logged to this directory."""

    vehicles: Dict[str, SmartVehicle] = field(default_factory=dict, init=False)
    """Vehicles associated with the account."""

    def __post_init__(self, password, log_responses):
        """Initialize the account."""

        if self.config is None:
            self.config = SmartClientConfiguration(
                SmartAuthentication(self.username, password),
                log_responses=log_responses,
            )

    async def login(self, force_refresh: bool = False) -> None:
        """Get the vehicles associated with the account."""
        if force_refresh is None:
            self.config.authentication = None
        await self.config.authentication.login()

    async def _init_vehicles(self) -> None:
        """Initialize vehicles from Smart servers."""
        _LOGGER.debug("Getting initial vehicle list")

        fetched_at = datetime.datetime.now(datetime.timezone.utc)

        async with SmartClient(self.config) as client:
            params = {
                "needSharedCar": 1,
                "userId": self.config.authentication.api_user_id,
            }
            for retry in range(2):
                try:
                    vehicles_response = await client.get(
                        API_BASE_URL + API_CARS_URL + "?" + utils.join_url_params(params),
                        headers={
                            **utils.generate_default_header(
                                client.config.authentication.device_id,
                                client.config.authentication.api_access_token,
                                params=params,
                                method="GET",
                                url=API_CARS_URL,
                            )
                        },
                    )
                    _LOGGER.debug(f"Got response {vehicles_response.status_code} from {vehicles_response.text}")
                except SmartTokenRefreshNecessary:
                    _LOGGER.debug(f"Got Token Error, retry: {retry}")
                    continue
                except SmartHumanCarConnectionError:
                    _LOGGER.debug(f"Got Human Car Connection Error, retry: {retry}")
                    self._init_vehicles()
                    continue
                break

            for vehicle in vehicles_response.json()["data"]["list"]:
                _LOGGER.debug(f"Found vehicle {vehicle}")
                self.add_vehicle(vehicle, fetched_at)

    def add_vehicle(self, vehicle, fetched_at):
        """Add a vehicle to the account."""
        self.vehicles[vehicle.get("vin")] = SmartVehicle(self, vehicle, fetched_at=fetched_at)

    async def get_vehicles(self, force_init: bool = False) -> None:
        """Get the vehicles associated with the account."""
        if self.config.authentication.api_user_id is None:
            await self.config.authentication.login()

        _LOGGER.debug("Getting vehicles for account %s", self.username)

        if len(self.vehicles) == 0 or force_init:
            await self._init_vehicles()

        for vin, vehicle in self.vehicles.items():
            _LOGGER.debug(f"Getting vehicle {vehicle.data}")
            await self.select_active_vehicle(vin)
            vehicle_info = await self.get_vehicle_information(vin)
            vehicle_soc = await self.get_vehicle_soc(vin)
            vehicle.combine_data(vehicle_info, charging_settings=vehicle_soc)

    async def select_active_vehicle(self, vin) -> None:
        """Select the active vehicle."""
        _LOGGER.debug(f"Selecting vehicle {vin}")
        data = json.dumps(
            {
                "vin": vin,
                "sessionToken": self.config.authentication.api_access_token,
                "language": "",
            }
        )
        async with SmartClient(self.config) as client:
            for retry in range(2):
                try:
                    r_car_info = await client.post(
                        API_BASE_URL + API_SELECT_CAR_URL,
                        headers={
                            **utils.generate_default_header(
                                client.config.authentication.device_id,
                                client.config.authentication.api_access_token,
                                params={},
                                method="POST",
                                url=API_SELECT_CAR_URL,
                                body=data,
                            )
                        },
                        data=data,
                    )
                    _LOGGER.debug(f"Got response {r_car_info.status_code} from {r_car_info.text}")
                except SmartTokenRefreshNecessary:
                    _LOGGER.debug(f"Got Token Error, retry: {retry}")
                    continue
                except SmartHumanCarConnectionError:
                    _LOGGER.debug(f"Got Human Car Connection Error, retry: {retry}")
                    self.select_active_vehicle(vin)
                    continue
                break

    async def get_vehicle_information(self, vin) -> str:
        """Get information about a vehicle."""
        _LOGGER.debug(f"Getting information for vehicle {vin}")
        params = {
            "latest": True,
            "target": "basic%2Cmore",
            "userId": self.config.authentication.api_user_id,
        }
        data = {}
        async with SmartClient(self.config) as client:
            for retry in range(2):
                try:
                    r_car_info = await client.get(
                        API_BASE_URL + "/remote-control/vehicle/status/" + vin + "?" + utils.join_url_params(params),
                        headers={
                            **utils.generate_default_header(
                                client.config.authentication.device_id,
                                client.config.authentication.api_access_token,
                                params=params,
                                method="GET",
                                url="/remote-control/vehicle/status/" + vin,
                            )
                        },
                    )
                    _LOGGER.debug(f"Got response {r_car_info.status_code} from {r_car_info.text}")
                    self.vehicles.get(vin).combine_data(r_car_info.json()["data"])
                    data = r_car_info.json()["data"]
                except SmartTokenRefreshNecessary:
                    _LOGGER.debug(f"Got Token Error, retry: {retry}")
                    continue
                except SmartHumanCarConnectionError:
                    _LOGGER.debug(f"Got Human Car Connection Error, retry: {retry}")
                    self.select_active_vehicle(vin)
                    continue
                break
            if retry > 1:
                raise SmartAuthError("Could not get vehicle information")
        return data

    async def get_vehicle_soc(self, vin) -> str:
        """Get information about a vehicle."""
        _LOGGER.debug(f"Getting information for vehicle {vin}")
        params = {
            "setting": "charging",
        }
        data = {}
        async with SmartClient(self.config) as client:
            for retry in range(2):
                try:
                    r_car_info = await client.get(
                        API_BASE_URL
                        + "/remote-control/vehicle/status/soc/"
                        + vin
                        + "?"
                        + utils.join_url_params(params),
                        headers={
                            **utils.generate_default_header(
                                client.config.authentication.device_id,
                                client.config.authentication.api_access_token,
                                params=params,
                                method="GET",
                                url="/remote-control/vehicle/status/soc/" + vin,
                            )
                        },
                    )
                    _LOGGER.debug(f"Got response {r_car_info.status_code} from {r_car_info.text}")
                    self.vehicles.get(vin).combine_data(r_car_info.json()["data"])
                    data = r_car_info.json()["data"]
                except SmartTokenRefreshNecessary:
                    _LOGGER.debug(f"Got Token Error, retry: {retry}")
                    continue
                except SmartHumanCarConnectionError:
                    _LOGGER.debug(f"Got Human Car Connection Error, retry: {retry}")
                    self.select_active_vehicle(vin)
                    continue
                break
            if retry > 1:
                raise SmartAuthError("Could not get vehicle information")
        return data



import pytest
import respx

from pysmarthashtag.control.climate import HeatingLocation
from pysmarthashtag.tests.conftest import prepare_account_with_vehicles


@pytest.mark.asyncio
async def test_enable_climate(smart_fixture: respx.Router):
    """Test the set_climate_conditioning method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    await account.get_vehicle_information("TestVIN0000000001")
    assert account.vehicles["TestVIN0000000001"].climate_control
    climate_ctrl = account.vehicles["TestVIN0000000001"].climate_control
    result = await climate_ctrl.set_climate_conditioning(20, True)
    assert result


@pytest.mark.asyncio
async def test_disable_climate(smart_fixture: respx.Router):
    """Test the set_climate_conditioning method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    await account.get_vehicle_information("TestVIN0000000001")
    assert account.vehicles["TestVIN0000000001"].climate_control
    climate_ctrl = account.vehicles["TestVIN0000000001"].climate_control
    result = await climate_ctrl.set_climate_conditioning(20, False)
    assert result


@pytest.mark.asyncio
async def test_enable_climate_invalid_temperature(smart_fixture: respx.Router):
    """Test the set_climate_conditioning method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    await account.get_vehicle_information("TestVIN0000000001")
    assert account.vehicles["TestVIN0000000001"].climate_control
    climate_ctrl = account.vehicles["TestVIN0000000001"].climate_control
    with pytest.raises(ValueError) as excinfo:
        await climate_ctrl.set_climate_conditioning(-20, True)
    assert str(excinfo.value) == "Temperature must be between 16 and 30 degrees."


@pytest.mark.asyncio
async def test_enable_seatheating(smart_fixture: respx.Router):
    """Test the set_climate_conditioning method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    await account.get_vehicle_information("TestVIN0000000001")
    assert account.vehicles["TestVIN0000000001"].climate_control
    climate_ctrl = account.vehicles["TestVIN0000000001"].climate_control
    for loc in HeatingLocation:
        climate_ctrl.set_heating_level(loc, 3)
    result = await climate_ctrl.set_climate_conditioning(20, True)
    assert result


@pytest.mark.asyncio
async def test_enable_seatheating_invalid_level(smart_fixture: respx.Router):
    """Test the set_climate_conditioning method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    await account.get_vehicle_information("TestVIN0000000001")
    assert account.vehicles["TestVIN0000000001"].climate_control
    climate_ctrl = account.vehicles["TestVIN0000000001"].climate_control
    with pytest.raises(ValueError) as excinfo:
        climate_ctrl.set_heating_level(HeatingLocation.DRIVER_SEAT, 4)
    assert str(excinfo.value) == "Seat heating level must be between 0 and 3."



"""Mock for API Backend."""

import json
from pathlib import Path
from typing import Any, Union

TEST_USERNAME = "some_user"
TEST_PASSWORD = "my_secret"


RESPONSE_DIR = Path(__file__).parent / "replys"


def load_response(path: Union[Path, str]) -> Any:
    """Load a stored response."""
    with open(path, "rb") as file:
        if Path(path).suffix == ".json":
            return json.load(file)
        return file.read().decode("UTF-8")



"""Fixtures for Smart tests."""

import respx

from pysmarthashtag.const import (
    API_BASE_URL,
    API_CARS_URL,
    API_SELECT_CAR_URL,
    API_SESION_URL,
    AUTH_URL,
    CONTEXT_URL,
    LOGIN_URL,
)
from pysmarthashtag.tests import RESPONSE_DIR, load_response


class SmartMockRouter(respx.MockRouter):
    """Stateful MockRouter for Smart APIs."""

    def __init__(
        self,
    ) -> None:
        """Initialize the SmartMockRouter with clean responses."""
        super().__init__(assert_all_called=False)

        self.add_login_routes()

    # # # # # # # # # # # # # # # # # # # # # # # #
    # Routes
    # # # # # # # # # # # # # # # # # # # # # # # #

    def add_login_routes(self) -> None:
        """Add routes for login."""

        # Login context
        self.get(CONTEXT_URL).respond(302, headers={"location": load_response(RESPONSE_DIR / "auth_context.url")})
        self.get(load_response(RESPONSE_DIR / "auth_context.url")).respond(
            200,
        )
        self.post(LOGIN_URL).respond(
            200,
            json={"sessionInfo": {"login_token": "TestToken", "expires_in": 3600}},
            headers={"location": load_response(RESPONSE_DIR / "auth_intermediate.url")},
        )
        self.get(
            AUTH_URL + "?context=eu1_tk1.WTygkw2cU6QLANxlBVba0lP4ndHARZgp14RhsVBRiVE.1705656141&login_token=TestToken"
        ).respond(302, headers={"location": "https://auth.smart.com" + load_response(RESPONSE_DIR / "auth_result.url")})
        self.get("https://auth.smart.com" + load_response(RESPONSE_DIR / "auth_result.url")).respond(
            200,
            json=load_response(RESPONSE_DIR / "login_result.json"),
            headers={"location": load_response(RESPONSE_DIR / "auth_result.url")},
        )
        self.post(API_BASE_URL + API_SESION_URL + "?identity_type=smart").respond(
            200,
            json=load_response(RESPONSE_DIR / "api_access.json"),
        )
        self.get(API_BASE_URL + API_CARS_URL + "?needSharedCar=1&userId=112233").respond(
            200,
            json=load_response(RESPONSE_DIR / "vehicle_response.json"),
        )
        self.post(API_BASE_URL + API_SELECT_CAR_URL).respond(200, json={})
        self.get(
            API_BASE_URL
            + "/remote-control/vehicle/status/TestVIN0000000001?latest=True&target=basic%2Cmore&userId=112233"
        ).respond(
            200,
            json=load_response(RESPONSE_DIR / "vehicle_info.json"),
        )
        self.get(
            API_BASE_URL
            + "/remote-control/vehicle/status/TestVIN0000000002?latest=True&target=basic%2Cmore&userId=112233"
        ).respond(
            200,
            json=load_response(RESPONSE_DIR / "vehicle_info_dc_charging.json"),
        )
        self.put(API_BASE_URL + "/remote-control/vehicle/telematics/TestVIN0000000001").respond(
            200, json=load_response(RESPONSE_DIR / "climate_success.json")
        )
        self.get(API_BASE_URL + "/remote-control/vehicle/status/soc/TestVIN0000000001?setting=charging").respond(
            200,
            json=load_response(RESPONSE_DIR / "soc_90.json"),
        )
        self.get(API_BASE_URL + "/remote-control/vehicle/status/soc/TestVIN0000000002?setting=charging").respond(
            200,
            json=load_response(RESPONSE_DIR / "soc_90.json"),
        )



"""Safety models for pysmarthashtag."""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from pysmarthashtag.models import VehicleDataBase

_LOGGER = logging.getLogger(__name__)


@dataclass
class Safety(VehicleDataBase):
    """Provides an accessible version of the vehicle's safety data."""

    central_locking_status: Optional[int] = None
    """The state of the central locking."""

    door_lock_status_driver: Optional[int] = None
    """The state of the driver's door lock."""

    door_lock_status_driver_rear: Optional[int] = None
    """The state of the driver's rear door lock."""

    door_lock_status_passenger: Optional[int] = None
    """The state of the passenger's door lock."""

    door_lock_status_passenger_rear: Optional[int] = None
    """The state of the passenger's rear door lock."""

    door_open_status_driver: Optional[int] = None
    """The state of the driver's door."""

    door_open_status_driver_rear: Optional[int] = None
    """The state of the driver's rear door."""

    door_open_status_passenger: Optional[int] = None
    """The state of the passenger's door."""

    door_open_status_passenger_rear: Optional[int] = None
    """The state of the passenger's rear door."""

    door_pos_driver: Optional[int] = None
    """The position of the driver's door."""

    door_pos_driver_rear: Optional[int] = None
    """The position of the driver's rear door."""

    door_pos_passenger: Optional[int] = None
    """The position of the passenger's door."""

    door_pos_passenger_rear: Optional[int] = None
    """The position of the passenger's rear door."""

    electric_park_brake_status: Optional[int] = None
    """The state of the electric park brake."""

    engine_hood_open_status: Optional[int] = None
    """The state of the engine hood."""

    seat_belt_status_driver: Optional[bool] = None
    """The state of the driver's seat belt."""

    seat_belt_status_driver_rear: Optional[bool] = None
    """The state of the driver's rear seat belt."""

    seat_belt_status_mid_rear: Optional[bool] = None
    """The state of the middle rear seat belt."""

    seat_belt_status_passenger: Optional[bool] = None
    """The state of the passenger's seat belt."""

    seat_belt_status_passenger_rear: Optional[bool] = None
    """The state of the passenger's rear seat belt."""

    seat_belt_status_th_driver_rear: Optional[bool] = None
    """The state of the driver's rear seat belt."""

    seat_belt_status_th_passenger_rear: Optional[bool] = None
    """The state of the passenger's rear seat belt."""

    srs_crash_status: Optional[int] = None
    """The state of the SRS crash."""

    tank_flap_status: Optional[int] = None
    """The state of the tank flap."""

    trunk_lock_status: Optional[int] = None
    """The state of the trunk lock."""

    trunk_open_status: Optional[int] = None
    """The state of the trunk."""

    vehicle_alarm: Optional[Dict] = None
    """The state of the vehicle alarm."""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the battery data based on Ids."""
        _LOGGER.debug(f"Parsing battery data: {vehicle_data}")
        if "vehicleStatus" not in vehicle_data:
            return None
        retval: Dict[str, Any] = {}
        try:
            evStatus = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["drivingSafetyStatus"]

            retval["central_locking_status"] = int(evStatus.get("centralLockingStatus"))
            retval["door_lock_status_driver"] = int(evStatus.get("doorLockStatusDriver"))
            retval["door_lock_status_driver_rear"] = int(evStatus.get("doorLockStatusDriverRear"))
            retval["door_lock_status_passenger"] = int(evStatus.get("doorLockStatusPassenger"))
            retval["door_lock_status_passenger_rear"] = int(evStatus.get("doorLockStatusPassengerRear"))
            retval["door_open_status_driver"] = int(evStatus.get("doorOpenStatusDriver"))
            retval["door_open_status_driver_rear"] = int(evStatus.get("doorOpenStatusDriverRear"))
            retval["door_open_status_passenger"] = int(evStatus.get("doorOpenStatusPassenger"))
            retval["door_open_status_passenger_rear"] = int(evStatus.get("doorOpenStatusPassengerRear"))
            retval["door_pos_driver"] = int(evStatus.get("doorPosDriver"))
            retval["door_pos_driver_rear"] = int(evStatus.get("doorPosDriverRear"))
            retval["door_pos_passenger"] = int(evStatus.get("doorPosPassenger"))
            retval["door_pos_passenger_rear"] = int(evStatus.get("doorPosPassengerRear"))
            retval["electric_park_brake_status"] = int(evStatus.get("electricParkBrakeStatus"))
            retval["engine_hood_open_status"] = int(evStatus.get("engineHoodOpenStatus"))
            retval["seat_belt_status_driver"] = bool(evStatus.get("seatBeltStatusDriver"))
            retval["seat_belt_status_driver_rear"] = bool(evStatus.get("seatBeltStatusDriverRear"))
            retval["seat_belt_status_mid_rear"] = bool(evStatus.get("seatBeltStatusMidRear"))
            retval["seat_belt_status_passenger"] = bool(evStatus.get("seatBeltStatusPassenger"))
            retval["seat_belt_status_passenger_rear"] = bool(evStatus.get("seatBeltStatusPassengerRear"))
            retval["seat_belt_status_th_driver_rear"] = bool(evStatus.get("seatBeltStatusThDriverRear"))
            retval["seat_belt_status_th_passenger_rear"] = bool(evStatus.get("seatBeltStatusThPassengerRear"))
            retval["srs_crash_status"] = int(evStatus.get("srsCrashStatus"))
            retval["tank_flap_status"] = int(evStatus.get("tankFlapStatus"))
            retval["trunk_lock_status"] = int(evStatus.get("trunkLockStatus"))
            retval["trunk_open_status"] = int(evStatus.get("trunkOpenStatus"))
            retval["vehicle_alarm"] = evStatus.get("vehicleAlarm")

            retval["timestamp"] = datetime.fromtimestamp(int(vehicle_data["vehicleStatus"]["updateTime"]) / 1000)
        except KeyError as e:
            _LOGGER.warning(f"Safety info not available: {e}")
        finally:
            return retval



"""Battery models for pysmarthashtag."""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from pysmarthashtag.models import ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


@dataclass
class Maintenance(VehicleDataBase):
    """Provides an accessible version of the vehicle's maintenance data."""

    days_to_service: Optional[int] = None
    """Days to service."""

    engine_hours_to_service: Optional[int] = None
    """Engine hours to service."""

    odometer: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Odometer."""

    break_fluid_level_status: Optional[int] = None
    """Break fluid level status."""

    main_battery_state_of_charge: Optional[int] = None
    """Main battery state of charge."""

    main_battery_charge_level: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Main battery charge level."""

    main_battery_energy_level: Optional[int] = None
    """Main battery energy level."""

    main_battery_state_of_health: Optional[int] = None
    """Main battery state of health."""

    main_batter_power_level: Optional[int] = None
    """Main battery power level."""

    main_battery_voltage: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Main battery voltage."""

    distance_to_service: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Distance to service."""

    service_warning_status: Optional[int] = None
    """Service warning status."""

    washer_fluid_level_status: Optional[int] = None
    """Washer fluid level status."""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the battery data based on Ids."""
        if "vehicleStatus" not in vehicle_data:
            return None
        retval: Dict[str, Any] = {}
        try:
            evStatus = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["maintenanceStatus"]
            _LOGGER.debug(f"Parsing maintenance data: {evStatus}")
            retval["main_battery_state_of_charge"] = int(evStatus["mainBatteryStatus"]["stateOfCharge"])
            retval["main_battery_charge_level"] = ValueWithUnit(
                float(evStatus["mainBatteryStatus"]["chargeLevel"]), "%"
            )
            retval["main_battery_energy_level"] = int(evStatus["mainBatteryStatus"]["energyLevel"])
            retval["main_battery_state_of_health"] = int(evStatus["mainBatteryStatus"]["stateOfHealth"])
            retval["main_batter_power_level"] = int(evStatus["mainBatteryStatus"]["powerLevel"])
            retval["main_battery_voltage"] = ValueWithUnit(float(evStatus["mainBatteryStatus"]["voltage"]), "V")

            retval["odometer"] = ValueWithUnit(int(float(evStatus["odometer"])), "km")
            retval["days_to_service"] = int(evStatus["daysToService"])
            retval["distance_to_service"] = ValueWithUnit(int(evStatus["distanceToService"]), "km")
            retval["service_warning_status"] = int(evStatus["serviceWarningStatus"])
            retval["break_fluid_level_status"] = int(evStatus["brakeFluidLevelStatus"])
            retval["washer_fluid_level_status"] = int(evStatus["washerFluidLevelStatus"])
            retval["engine_hours_to_service"] = int(evStatus["engineHrsToService"])

            retval["timestamp"] = datetime.fromtimestamp(int(vehicle_data["vehicleStatus"]["updateTime"]) / 1000)
        except KeyError as e:
            _LOGGER.warning(f"Battery info not available: {e}")
        finally:
            return retval



"""Battery models for pysmarthashtag."""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from pysmarthashtag.models import ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


@dataclass
class Running(VehicleDataBase):
    """Provides an accessible version of the vehicle's running data."""

    ahbc_status: Optional[int] = None
    """Adaptive high beam control status."""

    goodbye: Optional[int] = None
    """Goodbye Light."""

    home_safe: Optional[int] = None
    """Home Safe Light."""

    corner_light: Optional[int] = None
    """Corner light."""

    front_fog_light: Optional[int] = None
    """Front Fog light."""

    stop_light: Optional[int] = None
    """Stop light."""

    trip_meter1: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Trip meter 1."""

    trip_meter2: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Trip meter 2."""

    approach: Optional[int] = None
    """Approach light."""

    high_beam: Optional[int] = None
    """High beam light."""

    engine_coolant_level_status: Optional[int] = None
    """Engine coolant level status."""

    low_beam: Optional[int] = None
    """Low beam light."""

    position_light_rear: Optional[int] = None
    """Position light rear."""

    light_show: Optional[int] = None
    """Light show."""

    welcome: Optional[int] = None
    """Welcome light."""

    drl: Optional[int] = None
    """Daytime running light."""

    ahl: Optional[int] = None
    """Adaptive headlight."""

    trun_indicator_left: Optional[int] = None
    """Turn indicator left."""

    trun_indicator_right: Optional[int] = None
    """Turn indicator right."""

    adaptive_front_light: Optional[int] = None
    """Adaptive front lighting system."""

    dbl: Optional[int] = None
    """Double light."""

    average_speed: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Average speed."""

    position_light_front: Optional[int] = None
    """Position light front."""

    reverse_light: Optional[int] = None
    """Reverse light."""

    highway_light: Optional[int] = None
    """Highway light."""

    rear_fog_light: Optional[int] = None
    """Rear fog light."""

    flash_light: Optional[int] = None
    """Flash light."""

    all_weather_light: Optional[int] = None
    """All weather light."""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the battery data based on Ids."""
        if "vehicleStatus" not in vehicle_data:
            return None
        retval: Dict[str, Any] = {}
        try:
            evStatus = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["runningStatus"]
            _LOGGER.debug(f"Parsing running data: {evStatus}")
            retval["ahbc_status"] = int(evStatus.get("ahbc"))
            retval["goodbye"] = int(evStatus.get("goodbye"))
            retval["home_safe"] = int(evStatus.get("homeSafe"))
            retval["corner_light"] = int(evStatus.get("cornrgLi"))
            retval["front_fog_light"] = int(evStatus.get("frntFog"))
            retval["stop_light"] = int(evStatus.get("stopLi"))
            retval["trip_meter1"] = ValueWithUnit(float(evStatus.get("tripMeter1")), "km")
            retval["trip_meter2"] = ValueWithUnit(float(evStatus.get("tripMeter2")), "km")
            retval["approach"] = int(evStatus.get("approach"))
            retval["high_beam"] = int(evStatus.get("hiBeam"))
            retval["engine_coolant_level_status"] = int(evStatus.get("engineCoolantLevelStatus"))
            retval["low_beam"] = int(evStatus.get("loBeam"))
            retval["position_light_rear"] = int(evStatus.get("posLiRe"))
            retval["light_show"] = int(evStatus.get("ltgShow"))
            retval["welcome"] = int(evStatus.get("welcome"))
            retval["drl"] = int(evStatus.get("drl"))
            retval["ahl"] = int(evStatus.get("ahl"))
            retval["trun_indicator_left"] = int(evStatus.get("trunIndrLe"))
            retval["trun_indicator_right"] = int(evStatus.get("trunIndrRi"))
            retval["adaptive_front_light"] = int(evStatus.get("afs"))
            retval["dbl"] = int(evStatus.get("dbl"))
            retval["average_speed"] = ValueWithUnit(float(evStatus.get("avgSpeed")), "km/h")
            retval["position_light_front"] = int(evStatus.get("posLiFrnt"))
            retval["reverse_light"] = int(evStatus.get("reverseLi"))
            retval["highway_light"] = int(evStatus.get("hwl"))
            retval["rear_fog_light"] = int(evStatus.get("reFog"))
            retval["flash_light"] = int(evStatus.get("flash"))
            retval["all_weather_light"] = int(evStatus.get("allwl"))

            retval["timestamp"] = datetime.fromtimestamp(int(vehicle_data["vehicleStatus"]["updateTime"]) / 1000)
        except KeyError as e:
            _LOGGER.warning(f"Running info not available: {e}")
        finally:
            return retval



"""Tire models for pysmarthashtag."""

import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from pysmarthashtag.models import ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)

"""Charging state of electric vehicle."""


class TireLocation(Enum):
    """Enumeration of tire locations."""

    DRIVER = 0
    DRIVER_REAR = 1
    PASSANGER = 2
    PASSENAGER_REAR = 3


@dataclass
class Tires(VehicleDataBase):
    """Provides an accessible version of the vehicle's battery data."""

    temperature: Optional[List[ValueWithUnit]] = None
    """Temperature of the tires."""

    temperature_warning: Optional[List[bool]] = None
    """Temperature warning of the tires."""

    temperature_pre_warning: Optional[List[bool]] = None
    """Temperature pre warning of the tires."""

    tire_pressure: Optional[List[ValueWithUnit]] = None
    """Temperature status of the tires."""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the battery data based on Ids."""
        _LOGGER.debug(f"Parsing battery data: {vehicle_data}")
        retval: Dict[str, Any] = {}
        try:
            maintenance_status = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["maintenanceStatus"]
            retval["temperature"] = [
                ValueWithUnit(float(maintenance_status["tyreTempDriver"]), "C"),
                ValueWithUnit(float(maintenance_status["tyreTempDriverRear"]), "C"),
                ValueWithUnit(float(maintenance_status["tyreTempPassenger"]), "C"),
                ValueWithUnit(float(maintenance_status["tyreTempPassengerRear"]), "C"),
            ]
            retval["temperature_pre_warning"] = [
                True if maintenance_status["tyrePreWarningDriver"] == "1" else False,
                True if maintenance_status["tyrePreWarningDriverRear"] == "1" else False,
                True if maintenance_status["tyrePreWarningPassenger"] == "1" else False,
                True if maintenance_status["tyrePreWarningPassengerRear"] == "1" else False,
            ]
            retval["tire_pressure"] = [
                ValueWithUnit(float(maintenance_status["tyreStatusDriver"]), "kPa"),
                ValueWithUnit(float(maintenance_status["tyreStatusDriverRear"]), "kPa"),
                ValueWithUnit(float(maintenance_status["tyreStatusPassenger"]), "kPa"),
                ValueWithUnit(float(maintenance_status["tyreStatusPassengerRear"]), "kPa"),
            ]
            retval["temperature_warning"] = [
                True if maintenance_status["tyreTempWarningDriver"] == "1" else False,
                True if maintenance_status["tyreTempWarningDriverRear"] == "1" else False,
                True if maintenance_status["tyreTempWarningPassenger"] == "1" else False,
                True if maintenance_status["tyreTempWarningPassengerRear"] == "1" else False,
            ]

        except KeyError as e:
            _LOGGER.debug(f"Tire info not available: {e}")
        finally:
            return retval



#!/usr/bin/python3
"""Simple executable to demonstrate and test the usage of the library."""

import argparse
import asyncio
import logging.config
import os
import time

from pysmarthashtag.account import SmartAccount
from pysmarthashtag.control.climate import HeatingLocation


def environ_or_required(key):
    """Return default or required argument based on the existence of an environment variable."""
    return {"default": os.environ.get(key)} if os.environ.get(key) else {"required": True}


def main_parser() -> argparse.ArgumentParser:
    """Create argument parser."""

    LOGGING_CONFIG = {
        "version": 1,
        "handlers": {
            "default": {"class": "logging.StreamHandler", "formatter": "http", "stream": "ext://sys.stderr"},
            "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "http",
                "filename": "pysmarthashtag.log",
                "maxBytes": 1024 * 1024 * 10,
                "backupCount": 10,
            },
        },
        "formatters": {
            "http": {
                "format": "%(levelname)s [%(asctime)s] %(name)s - %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            }
        },
        "loggers": {
            "pysmarthashtag": {
                "handlers": ["default", "file"],
                "level": "INFO",
            },
            "httpx": {
                "handlers": ["default"],
                "level": "ERROR",
            },
        },
    }

    logging.config.dictConfig(LOGGING_CONFIG)

    parser = argparse.ArgumentParser(description="Smart API demo")
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    _ = subparsers.add_parser("status", help="Get status of vehicle.")

    _ = subparsers.add_parser("info", help="Get info of vehicle.")

    watch_parser = subparsers.add_parser("watch", help="Watch vehicle.")
    watch_parser.add_argument("-i", help="scan intervall", default=60, type=int)

    climate_parser = subparsers.add_parser("climate", help="Set climate of vehicle.")
    climate_parser.add_argument("--vin", help="VIN of vehicle", default=None)
    climate_parser.add_argument("--temp", help="Temperature", default=22)
    climate_parser.add_argument("--active", help="Active", action="store_true")

    seatheating_parser = subparsers.add_parser("seatheating", help="Set heating of seats in vehicle.")
    seatheating_parser.add_argument("--vin", help="VIN of vehicle", default=None)
    seatheating_parser.add_argument("--level", type=int, choices=[1, 2, 3], help="Heating level (1-3)", default=3)
    seatheating_parser.add_argument("--temp", help="Temperature", default=22)
    seatheating_parser.add_argument("--active", help="Active", action="store_true")

    _add_default_args(parser)
    parser.set_defaults(func=parse_command)

    return parser


async def parse_command(args) -> None:
    """Parse command."""
    if args.command == "status":
        await get_status(args)
    elif args.command == "info":
        await get_vehicle_information(args)
    elif args.command == "watch":
        await watch_car(args)
    elif args.command == "climate":
        await set_climate(args)
    elif args.command == "seatheating":
        await set_seatheating(args)
    else:
        raise NotImplementedError(f"Command {args.command} not implemented.")


async def get_status(args) -> None:
    """Get status of vehicle."""
    account = SmartAccount(args.username, args.password)
    await account.get_vehicles()

    for vin, vehicle in account.vehicles.items():
        print(f"VIN: {vin} -> {vehicle.engine_state}")


async def get_vehicle_information(args) -> None:
    """Get status of vehicle."""
    account = SmartAccount(args.username, args.password)
    await account.get_vehicles()

    for vin, vehicle in account.vehicles.items():
        car = await account.get_vehicle_information(vin)
        print(f"VIN: {vin}")
        print(f"{car}")


async def watch_car(args) -> None:
    """Get status of vehicle."""
    account = SmartAccount(args.username, args.password)
    await account.get_vehicles()

    while True:
        for vin, vehicle in account.vehicles.items():
            car = await account.get_vehicle_information(vin)
            print(car)
        time.sleep(args.i)


async def set_climate(args) -> None:
    """Set climate of vehicle."""
    account = SmartAccount(args.username, args.password)
    await account.get_vehicles()
    if not args.vin:
        args.vin = list(account.vehicles.keys())[0]
    await account.get_vehicle_information(args.vin)

    climate_ctrl = account.vehicles[args.vin].climate_control
    await climate_ctrl.set_climate_conditioning(args.temp, args.active)


async def set_seatheating(args) -> None:
    """Set heating of driver's seat in vehicle."""
    account = SmartAccount(args.username, args.password)
    await account.get_vehicles()
    if not args.vin:
        args.vin = list(account.vehicles.keys())[0]
    await account.get_vehicle_information(args.vin)

    climate_ctrl = account.vehicles[args.vin].climate_control
    climate_ctrl.set_heating_level(HeatingLocation.DRIVER_SEAT, args.level)
    await climate_ctrl.set_climate_conditioning(args.temp, args.active)


def _add_default_args(parser: argparse.ArgumentParser):
    """Add the default arguments username, password to the parser."""
    parser.add_argument("--username", help="Smart username", **environ_or_required("SMART_USERNAME"))
    parser.add_argument("--password", help="Smart password", **environ_or_required("SMART_PASSWORD"))


def main():
    """Get arguments from parser and run function in event loop."""
    parser = main_parser()
    args = parser.parse_args()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(args.func(args))


if __name__ == "__main__":
    main()



import logging

import pytest
import respx
from httpx import Request, Response

from pysmarthashtag.const import API_BASE_URL, API_SELECT_CAR_URL
from pysmarthashtag.models import ValueWithUnit
from pysmarthashtag.tests import RESPONSE_DIR, load_response
from pysmarthashtag.tests.conftest import prepare_account_with_vehicles

_LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_login(smart_fixture: respx.Router):
    """Test the login flow."""
    account = await prepare_account_with_vehicles()
    assert account is not None


@pytest.mark.asyncio
async def test_get_vehicles(smart_fixture: respx.Router):
    """Test the get_vehicles method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    vehicles = account.vehicles
    assert len(vehicles) == 2
    assert vehicles["TestVIN0000000001"].vin == "TestVIN0000000001"
    assert vehicles["TestVIN0000000002"].vin == "TestVIN0000000002"


@pytest.mark.asyncio
async def test_get_vehicles_token_expired(smart_fixture: respx.Router):
    """Test the get_vehicles method."""

    def switch_response(request: Request, route: respx.Route) -> Response:
        json_responsees = [
            "vehicle_info.json",
            "token_expired.json",
            "vehicle_info2.json",
        ]
        _LOGGER.warning("Switching response to %s", json_responsees[route.call_count])
        return Response(200, json=load_response(RESPONSE_DIR / json_responsees[route.call_count]))

    smart_fixture.get(
        API_BASE_URL + "/remote-control/vehicle/status/TestVIN0000000001?latest=True&target=basic%2Cmore&userId=112233"
    ).mock(side_effect=switch_response)
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    assert account.vehicles["TestVIN0000000001"].engine_state == "engine_off"
    # now the token is invalid, refetched and data is updated
    await account.get_vehicle_information("TestVIN0000000001")

    assert account.vehicles["TestVIN0000000001"].engine_state == "engine_running"


@pytest.mark.asyncio
async def test_no_human_car_connection(smart_fixture: respx.Router):
    """Test the get_vehicles method."""

    did_call_car_selection = 0

    def switch_response(request: Request, route: respx.Route) -> Response:
        nonlocal did_call_car_selection
        json_responsees = [
            "Human_and_vehicle_relationship_does_not_exist.json",
            "vehicle_info.json",
        ]
        _LOGGER.warning("Switching response to %s", json_responsees[did_call_car_selection])
        return Response(200, json=load_response(RESPONSE_DIR / json_responsees[did_call_car_selection]))

    def count_car_selection(request: Request, route: respx.Route) -> Response:
        nonlocal did_call_car_selection
        did_call_car_selection += 1
        return Response(200, json={})

    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None

    vehicle_status = smart_fixture.get(
        API_BASE_URL + "/remote-control/vehicle/status/TestVIN0000000001?latest=True&target=basic%2Cmore&userId=112233"
    ).mock(side_effect=switch_response)
    car_connection = smart_fixture.post(API_BASE_URL + API_SELECT_CAR_URL).mock(side_effect=count_car_selection)

    await account.get_vehicle_information("TestVIN0000000001")
    assert car_connection.call_count == 2  # 2 times for the connection refresh (this is new??)
    assert vehicle_status.call_count == 3  # 2 times for the token refresh and 1 time for the inital call


@pytest.mark.asyncio
async def test_get_vehicle_chargin_dc(smart_fixture: respx.Router):
    """Test the get_vehicles method."""
    account = await prepare_account_with_vehicles()
    assert account is not None
    assert account.vehicles is not None
    vehicles = account.vehicles
    assert len(vehicles) == 2
    assert vehicles["TestVIN0000000002"].battery.charging_status == "DC_CHARGING"
    assert vehicles["TestVIN0000000002"].battery.is_charger_connected
    assert vehicles["TestVIN0000000002"].battery.charging_current == ValueWithUnit(value=102.6, unit="A")



"""Battery models for pysmarthashtag."""

import logging
import math
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from pysmarthashtag.models import ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)

"""Charging state of electric vehicle."""
ChargingState = [
    "NOT_CHARGING",
    "DEFAULT",
    "CHARGING",
    "ERROR",
    "COMPLETE",
    "FULLY_CHARGED",
    "FINISHED_FULLY_CHARGED",
    "FINISHED_NOT_FULL",
    "INVALID",
    "PLUGGED_IN",
    "WAITING_FOR_CHARGING",
    "TARGET_REACHED",
    "UNKNOWN",
    "UNKNOWN",
    "UNKNOWN",
    "DC_CHARGING",
]

DcChargingVoltLevels = [
    370,
    374,
    377,
    381,
    385,
    389,
    392,
    396,
    400,
    403,
    407,
    408,
    408,
    409,
    409,
    410,
    410,
    411,
    411,
    412,
    412,
    413,
    413,
    413,
    413,
    414,
    414,
    415,
    416,
    416,
    416,
    416,
    416,
    416,
    417,
    417,
    415,
    416,
    416,
    416,
    417,
    417,
    417,
    416,
    416,
    417,
    417,
    417,
    416,
    417,
    418,
    418,
    417,
    417,
    418,
    419,
    420,
    419,
    419,
    420,
    422,
    423,
    424,
    425,
    426,
    427,
    428,
    429,
    430,
    431,
    433,
    434,
    435,
    436,
    437,
    437,
    439,
    440,
    441,
    440,
    440,
    441,
    443,
    444,
    445,
    446,
    448,
    449,
    450,
    451,
    453,
    454,
    455,
    456,
    458,
    459,
    460,
    461,
    463,
    464,
    465,
]


@dataclass
class Battery(VehicleDataBase):
    """Provides an accessible version of the vehicle's battery data."""

    remaining_range: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Remaining range of the vehicle."""

    remaining_range_at_full_charge: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Remaining range at full charge of the vehicle."""

    remaining_battery_percent: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Remaining battery percent of the vehicle."""

    charging_status: Optional[str] = None
    """Charging status of the vehicle as string."""

    charging_status_raw: Optional[int] = None
    """Charging status of the vehicle."""

    charger_connection_status: Optional[int] = None
    """Charger connection status of the vehicle."""

    is_charger_connected: bool = False
    """Is the charger connected to the vehicle."""

    charging_voltage: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Charging voltage of the vehicle."""

    charging_current: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Charging current of the vehicle."""

    charging_power: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Charging power of the vehicle."""

    charging_time_remaining: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Charging time remaining of the vehicle."""

    charging_target_soc: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Charging target state of charge."""

    average_power_consumption: Optional[ValueWithUnit] = ValueWithUnit(None, "W")
    """Current average consumption"""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the battery data based on Ids."""
        _LOGGER.debug(f"Parsing battery data: {vehicle_data}")
        retval: Dict[str, Any] = {}
        try:
            evStatus = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["electricVehicleStatus"]
            if "distanceToEmptyOnBatteryOnly" in evStatus:
                retval["remaining_range"] = ValueWithUnit(int(evStatus["distanceToEmptyOnBatteryOnly"]), "km")
            if "distanceToEmptyOnBattery100Soc" in evStatus:
                retval["remaining_range_at_full_charge"] = ValueWithUnit(
                    int(evStatus["distanceToEmptyOnBattery100Soc"]), "km"
                )
            if "chargeLevel" in evStatus:
                retval["remaining_battery_percent"] = ValueWithUnit(int(evStatus["chargeLevel"]), "%")
            if "chargerState" in evStatus:
                status = int(evStatus["chargerState"])
                retval["charging_status"] = ChargingState[status] if status < len(ChargingState) else "UNKNOWN"
                retval["charging_status_raw"] = int(evStatus["chargerState"])
                retval["is_charger_connected"] = (
                    retval["charging_status"] == "PLUGGED_IN"
                    or retval["charging_status"] == "CHARGING"
                    or retval["charging_status"] == "DC_CHARGING"
                    or retval["charging_status"] == "COMPLETE"
                )
            if "statusOfChargerConnection" in evStatus:
                retval["charger_connection_status"] = int(evStatus["statusOfChargerConnection"])

            if "dcChargeIAct" in evStatus and retval["charging_status"] == "DC_CHARGING":
                retval["charging_voltage"] = ValueWithUnit(
                    DcChargingVoltLevels[retval["remaining_battery_percent"].value], "V"
                )
                retval["charging_current"] = ValueWithUnit(abs(float(evStatus["dcChargeIAct"])), "A")
            elif "chargeUAct" in evStatus and "chargeIAct" in evStatus:
                retval["charging_voltage"] = ValueWithUnit(float(evStatus["chargeUAct"]), "V")
                retval["charging_current"] = ValueWithUnit(float(evStatus["chargeIAct"]), "A")

            if "chargeIAct" in evStatus and "chargeUAct" in evStatus:
                retval["charging_power"] = ValueWithUnit(
                    float(evStatus["chargeUAct"]) * float(evStatus["chargeIAct"])
                    if retval["charging_voltage"].value < 260
                    else float(evStatus["chargeUAct"]) * float(evStatus["chargeIAct"]) * math.sqrt(3),
                    "W",
                )

            if "timeToFullyCharged" in evStatus:
                retval["charging_time_remaining"] = (
                    ValueWithUnit(int(evStatus["timeToFullyCharged"]), "min")
                    if evStatus["timeToFullyCharged"] != "2047"
                    else None
                )

            if "averPowerConsumption" in evStatus:
                retval["average_power_consumption"] = ValueWithUnit(float(evStatus["averPowerConsumption"]), "W")
            if "vehicleStatus" in vehicle_data and "updateTime" in vehicle_data["vehicleStatus"]:
                retval["timestamp"] = datetime.fromtimestamp(int(vehicle_data["vehicleStatus"]["updateTime"]) / 1000)
            if "soc" in vehicle_data:
                retval["charging_target_soc"] = ValueWithUnit(float(vehicle_data["soc"]) / 10, "%")
        except KeyError as e:
            _LOGGER.debug(f"Battery info not available: {e}")
        except Exception as e:
            _LOGGER.error(f"Error parsing battery data: {e}")
        finally:
            return retval



"""Generals models used for pysmarthastag."""

import logging
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, NamedTuple, Optional, Union

_LOGGER = logging.getLogger(__name__)


class StrEnum(str, Enum):
    """Enumerate strings."""

    @classmethod
    def _missing_(cls, value):
        has_unknown = False
        for member in cls:
            if member.value.upper() == "UNKNOWN":
                has_unknown = True
            if member.value.upper() == value.upper():
                return member
        if has_unknown:
            _LOGGER.warning("Unknown value %s for enum %s", value, cls.__name__)
            return getattr(cls, "UNKNOWN")
        raise ValueError(f"{value} is not a valid {cls.__name__}")


@dataclass
class VehicleDataBase:
    """Base class for vehicle data."""

    timestamp: Optional[datetime] = None

    @classmethod
    def from_vehicle_data(cls, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = cls._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return cls(**parsed)
        return None

    def update_from_vehicle_data(self, vehicle_data: Dict):
        """Update the instance with data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        parsed.update(self._update_after_parse(vehicle_data))
        if len(parsed) > 0:
            self.__dict__.update(parsed)

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the vehicle data."""
        raise NotImplementedError()

    def _update_after_parse(self, parsed: Dict) -> Dict:
        """Update the instance with data from API."""
        return parsed


class ValueWithUnit(NamedTuple):
    """A value with a corresponding unit."""

    value: Optional[Union[int, float]]
    unit: Optional[str]


@dataclass
class AnonymizedResponse:
    """An anonymized response."""

    filename: str
    content: Optional[Union[List, Dict, str]] = None


class SmartAPIError(Exception):
    """General Smart web API error."""


class SmartAuthError(SmartAPIError):
    """Auth-related error from Smart web API (HTTP status codes 401 and 403)."""


class SmartTokenRefreshNecessary(SmartAPIError):
    """Token refresh is necessary (Response Code 1402)."""


class SmartHumanCarConnectionError(SmartAPIError):
    """Human and vehicle connection does not exist (Response Code 8006)."""


class SmartQuotaError(SmartAPIError):
    """Quota exceeded on Smart web API."""


class SmartRemoteServiceError(SmartAPIError):
    """Error when executing web services."""


def get_element_from_dict_maybe(
    data: dict, *path: str, default: "Any|None" = None
) -> Optional[Union[dict, str, int, float]]:
    """Get an element from a dict by path."""
    if len(path) == 0:
        return data
    if path[0] not in data:
        return default
    return get_element_from_dict_maybe(data[path[0]], *path[1:])



"""Position models for pysmarthashtag."""

import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional

from pysmarthashtag.models import ValueWithUnit, VehicleDataBase, get_element_from_dict_maybe

_LOGGER = logging.getLogger(__name__)

"""Position of electric vehicle."""


@dataclass
class Position(VehicleDataBase):
    """Provides an accessible version of the vehicle's battery data."""

    altitude: Optional[ValueWithUnit] = None
    """Altitude of the vehicle."""

    latitude: Optional[int] = None
    """Latitude of the vehicle."""

    longitude: Optional[int] = None
    """Longitude of the vehicle."""

    position_can_be_trusted: Optional[bool] = None
    """Position can be trusted."""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the position data based on Ids."""
        _LOGGER.debug(f"Parsing position data: {vehicle_data}")
        retval: Dict[str, Any] = {}
        position = get_element_from_dict_maybe(vehicle_data, "vehicleStatus", "basicVehicleStatus", "position")
        try:
            retval["altitude"] = ValueWithUnit(int(position["altitude"]), "m")
            retval["latitude"] = int(position["latitude"])
            retval["longitude"] = int(position["longitude"])
            retval["position_can_be_trusted"] = True if position["posCanBeTrusted"] == "true" else False

        except KeyError as e:
            _LOGGER.debug(f"Position info not available: {e}")
        finally:
            return retval



import base64
import hashlib
import hmac
import logging
import secrets
import time
from typing import Dict

_LOGGER = logging.getLogger(__name__)


def join_url_params(args: Dict) -> str:
    """Join params for adding to URL."""
    return "&".join([f"{key}={value}" for key, value in args.items()])


def _create_sign(nonce: str, params: Dict, timestamp: str, method: str, url: str, body=None) -> str:
    """Create a signature for the request."""
    md5sum = base64.b64encode(hashlib.md5(body.encode()).digest()).decode() if body else "1B2M2Y8AsgTpgAmY7PhCfg=="
    url_params = join_url_params(params)
    payload = f"""application/json;responseformat=3
x-api-signature-nonce:{nonce}
x-api-signature-version:1.0

{url_params}
{md5sum}
{timestamp}
{method}
{url}"""
    _LOGGER.debug("Payload: %s", payload)
    secret = base64.b64decode("NzRlNzQ2OWFmZjUwNDJiYmJlZDdiYmIxYjM2YzE1ZTk=")
    payload = payload.encode("utf-8")
    hashed = hmac.new(secret, payload, hashlib.sha1).digest()
    signature = base64.b64encode(hashed).decode()
    _LOGGER.debug("Signature: %s", signature)
    return signature


def generate_default_header(
    device_id: str, access_token: str, params: Dict, method: str, url: str, body=None
) -> Dict[str, str]:
    """Generate a header for HTTP requests to the server."""
    timestamp = create_correct_timestamp()
    nonce = secrets.token_hex(8)
    sign = _create_sign(nonce, params, timestamp, method, url, body)
    header = {
        "x-app-id": "SmartAPPEU",
        "accept": "application/json;responseformat=3",
        "x-agent-type": "iOS",
        "x-device-type": "mobile",
        "x-operator-code": "SMART",
        "x-device-identifier": device_id,
        "x-env-type": "production",
        "x-version": "smartNew",
        "accept-language": "en_US",
        "x-api-signature-version": "1.0",
        "x-api-signature-nonce": nonce,
        "x-device-manufacture": "Apple",
        "x-device-brand": "Apple",
        "x-device-model": "iPhone",
        "x-agent-version": "17.1",
        "content-type": "application/json; charset=utf-8",
        "user-agent": "Hello smart/1.4.0 (iPhone; iOS 17.1; Scale/3.00)",
        "x-signature": sign,
        "x-timestamp": str(timestamp),
    }
    if access_token:
        header["authorization"] = access_token

    _LOGGER.debug(
        f"Constructed Login: {join_url_params(params)} - {access_token} - {method} - {url} - {body} -> {header}"
    )
    return header


def create_correct_timestamp() -> str:
    """Create a correct timestamp for the request."""
    return str(int(time.time() * 1000))



import logging
from collections import defaultdict, deque
from dataclasses import dataclass
from typing import Deque, Optional

import httpx

from pysmarthashtag.api.authentication import SmartAuthentication
from pysmarthashtag.const import (
    HTTPX_TIMEOUT,
    SERVER_URL,
)
from pysmarthashtag.models import (
    AnonymizedResponse,
    SmartHumanCarConnectionError,
    SmartTokenRefreshNecessary,
)

_LOGGER = logging.getLogger(__name__)

RESPONSE_STORE: Deque[AnonymizedResponse] = deque(maxlen=10)


@dataclass
class SmartClientConfiguration:
    """Stores global settings for SmartClient."""

    authentication: SmartAuthentication
    log_responses: Optional[bool] = False

    def set_log_responses(self, log_responses: bool) -> None:
        """Set if responses are logged and clear response store."""

        self.log_responses = log_responses
        RESPONSE_STORE.clear()


class SmartClient(httpx.AsyncClient):
    """Async HTTP client based on `httpx.AsyncClient` with automated OAuth token refresh."""

    last_message: str = ""

    def __init__(self, config: SmartClientConfiguration, *args, **kwargs):
        self.config = config

        # Add authentication
        # kwargs["auth"] = self.config.authentication

        # Increase timeout
        kwargs["timeout"] = httpx.Timeout(HTTPX_TIMEOUT)

        # Set default values
        kwargs["base_url"] = kwargs.get("base_url") or SERVER_URL

        # Register event hooks
        kwargs["event_hooks"] = defaultdict(list, **kwargs.get("event_hooks", {}))

        async def log_request(request):
            if request.method == "POST":
                await request.aread()
                _LOGGER.debug(
                    f"Request: {request.method} {request.url} - {request.headers} - {request.content.decode()}"
                )
            else:
                _LOGGER.debug(f"Request: {request.method} {request.url} - {request.headers}")

        async def log_response(response):
            await response.aread()
            request = response.request
            _LOGGER.debug(f"Response: {request.method} {request.url} - Status {response.status_code}")

        kwargs["event_hooks"]["response"].append(log_response)
        kwargs["event_hooks"]["request"].append(log_request)

        # Event hook which calls raise_for_status on all requests
        async def raise_for_status_event_handler(response: httpx.Response):
            """Event handler that automatically raises HTTPStatusErrors when attached.

            Will read out response JSON for code and message
            """
            response_data = response.json()
            if "message" in response_data:
                self.last_message = response_data["message"]
            if "code" in response_data and response_data["code"] == "1402":
                await self.config.authentication.login()
                raise SmartTokenRefreshNecessary("Token expired, refresh token, do request again.")
            if "code" in response_data and response_data["code"] == "8006":
                raise SmartHumanCarConnectionError(
                    "Human and vehicle relationship does not exist, selct car and do request again."
                )
            elif "code" in response_data and response_data["code"] != "1000" and "message" in response_data:
                raise httpx.HTTPStatusError(
                    response=response,
                    request=response.request,
                    message=f"{response_data['code']}: {response_data['message']}",
                )

        kwargs["event_hooks"]["response"].append(raise_for_status_event_handler)

        super().__init__(*args, **kwargs)



"""Battery models for pysmarthashtag."""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from pysmarthashtag.models import ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


@dataclass
class Climate(VehicleDataBase):
    """Provides an accessible version of the vehicle's climate data."""

    air_blower_active: Optional[bool] = None
    """The state of the air blower."""

    cds_climate_active: Optional[bool] = None
    """The state of the climate control system."""

    climate_over_heat_protection_active: Optional[bool] = None
    """The state of the climate overheat protection."""

    curtain_open_status: Optional[bool] = None
    """The state of the curtail open status."""

    curtain_position: Optional[int] = None
    """The position of the curtain."""

    defrosting_active: Optional[bool] = None
    """The state of the defrosting."""

    driver_heating_detail: Optional[int] = None
    """The position of the driver's heating."""

    driver_heating_status: Optional[bool] = None
    """The state of the driver's heating."""

    driver_ventilation_detail: Optional[int] = None
    """The position of the driver's ventilation."""

    driver_ventilation_status: Optional[bool] = None
    """The state of the driver's ventilation."""

    exterior_temperature: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """The exterior temperature."""

    frag_active: Optional[bool] = None
    """The state of the frag."""

    interior_temperature: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """The interior temperature."""

    passenger_heating_detail: Optional[int] = None
    """The position of the passenger's heating."""

    passenger_heating_status: Optional[bool] = None
    """The state of the passenger's heating."""

    passenger_ventilation_detail: Optional[int] = None
    """The position of the passenger's ventilation."""

    passenger_ventilation_status: Optional[bool] = None
    """The state of the passenger's ventilation."""

    pre_climate_active: Optional[bool] = None
    """The state of the pre-climate."""

    rear_left_heating_detail: Optional[int] = None
    """The position of the left rear heating."""

    rear_left_heating_status: Optional[bool] = None
    """The state of the left rear heating."""

    rear_left_ventilation_detail: Optional[int] = None
    """The position of the left rear ventilation."""

    rear_left_ventilation_status: Optional[bool] = None
    """The state of the left rear ventilation."""

    rear_right_heating_detail: Optional[int] = None
    """The position of the right rear heating."""

    rear_right_heating_status: Optional[bool] = None
    """The state of the right rear heating."""

    rear_right_ventilation_detail: Optional[int] = None
    """The position of the right rear ventilation."""

    rear_right_ventilation_status: Optional[bool] = None
    """The state of the right rear ventilation."""

    steering_wheel_heating_status: Optional[bool] = None
    """The state of the steering wheel heating."""

    sun_curtain_rear_open_status: Optional[bool] = None
    """The state of the rear sun curtain."""

    sun_curtain_rear_position: Optional[int] = None
    """The position of the rear sun curtain."""

    sunroof_open_status: Optional[bool] = None
    """The state of the sunroof."""

    sunroof_position: Optional[int] = None
    """The position of the sunroof."""

    window_driver_position: Optional[int] = None
    """The position of the driver's window."""

    window_driver_rear_position: Optional[int] = None
    """The position of the rear driver's window."""

    window_passenger_position: Optional[int] = None
    """The position of the passenger's window."""

    window_passenger_rear_position: Optional[int] = None
    """The position of the rear passenger's window."""

    window_driver_status: Optional[bool] = None
    """The state of the driver's window."""

    window_driver_rear_status: Optional[bool] = None
    """The state of the rear driver's window."""

    window_passenger_status: Optional[bool] = None
    """The state of the passenger's window."""

    window_passenger_rear_status: Optional[bool] = None
    """The state of the rear passenger's window."""

    interior_PM25: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """The interior PM2.5 value."""

    relative_humidity: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """The relative humidity."""

    @classmethod
    def from_vehicle_data(self, vehicle_data: Dict):
        """Create a new instance based on data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return self(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(self, vehicle_data: Dict) -> Optional[Dict]:
        """Parse the battery data based on Ids."""
        _LOGGER.debug(f"Parsing battery data: {vehicle_data}")
        if "vehicleStatus" not in vehicle_data:
            return None
        retval: Dict[str, Any] = {}
        try:
            evStatus = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["climateStatus"]

            retval["air_blower_active"] = True if evStatus["airBlowerActive"] == "true" else False
            retval["cds_climate_active"] = True if evStatus["cdsClimateActive"] == "true" else False
            retval["climate_over_heat_protection_active"] = (
                True if evStatus["climateOverHeatProActive"] == "true" else False
            )
            retval["curtain_open_status"] = True if evStatus["curtainOpenStatus"] == "1" else False
            retval["curtain_position"] = int(evStatus["curtainPos"])
            retval["defrosting_active"] = True if evStatus["defrost"] == "true" else False
            retval["driver_heating_detail"] = int(evStatus["drvHeatDetail"])
            retval["driver_heating_status"] = True if evStatus["drvHeatSts"] == "true" else False
            retval["driver_ventilation_detail"] = int(evStatus["drvVentDetail"])
            retval["driver_ventilation_status"] = True if evStatus["drvVentSts"] == "true" else False
            retval["exterior_temperature"] = ValueWithUnit(float(evStatus["exteriorTemp"]), "°C")
            retval["frag_active"] = evStatus["fragActive"]
            retval["interior_temperature"] = ValueWithUnit(float(evStatus["interiorTemp"]), "°C")
            retval["passenger_heating_detail"] = int(evStatus["passHeatingDetail"])
            retval["passenger_heating_status"] = True if evStatus["passHeatingSts"] == "true" else False
            retval["passenger_ventilation_detail"] = int(evStatus["passVentDetail"])
            retval["passenger_ventilation_status"] = True if evStatus["passVentSts"] == "true" else False
            retval["pre_climate_active"] = evStatus["preClimateActive"]
            retval["rear_left_heating_detail"] = int(evStatus["rlHeatingDetail"])
            retval["rear_left_heating_status"] = True if evStatus["rlHeatingSts"] == "true" else False
            retval["rear_left_ventilation_detail"] = int(evStatus["rlVentDetail"])
            retval["rear_left_ventilation_status"] = True if evStatus["rlVentSts"] == "true" else False
            retval["rear_right_heating_detail"] = int(evStatus["rrHeatingDetail"])
            retval["rear_right_heating_status"] = True if evStatus["rrHeatingSts"] == "true" else False
            retval["rear_right_ventilation_detail"] = int(evStatus["rrVentDetail"])
            retval["rear_right_ventilation_status"] = True if evStatus["rrVentSts"] == "true" else False
            retval["steering_wheel_heating_status"] = True if evStatus["steerWhlHeatingSts"] == "true" else False
            retval["sun_curtain_rear_open_status"] = True if evStatus["sunCurtainRearOpenStatus"] == "true" else False
            retval["sun_curtain_rear_position"] = int(evStatus["sunCurtainRearPos"])
            retval["sunroof_open_status"] = True if evStatus["sunroofOpenStatus"] == "true" else False
            retval["sunroof_position"] = int(evStatus["sunroofPos"])
            retval["window_driver_position"] = int(evStatus["winPosDriver"])
            retval["window_driver_rear_position"] = int(evStatus["winPosDriverRear"])
            retval["window_passenger_position"] = int(evStatus["winPosPassenger"])
            retval["window_passenger_rear_position"] = int(evStatus["winPosPassengerRear"])
            retval["window_driver_status"] = int(evStatus["winStatusDriver"])
            retval["window_driver_rear_status"] = int(evStatus["winStatusDriverRear"])
            retval["window_passenger_status"] = int(evStatus["winStatusPassenger"])
            retval["window_passenger_rear_status"] = int(evStatus["winStatusPassengerRear"])

            evStatus = vehicle_data["vehicleStatus"]["additionalVehicleStatus"]["pollutionStatus"]

            retval["interior_PM25"] = ValueWithUnit(float(evStatus["interiorPM25"]), "μg/m³")
            retval["relative_humidity"] = ValueWithUnit(float(evStatus["relHumSts"]), "%")

            retval["timestamp"] = datetime.fromtimestamp(int(vehicle_data["vehicleStatus"]["updateTime"]) / 1000)
        except KeyError as e:
            _LOGGER.warning(f"Climate info not available: {e}")
        finally:
            return retval



"""Authentication management for Smart APIs."""

import asyncio
import datetime
import json
import logging
import math
import secrets
from collections import defaultdict
from typing import AsyncGenerator, Generator, Optional

import httpx
from httpx._models import Request, Response

from pysmarthashtag.api import utils
from pysmarthashtag.const import (
    API_BASE_URL,
    API_KEY,
    API_SESION_URL,
    AUTH_URL,
    CONTEXT_URL,
    HTTPX_TIMEOUT,
    LOGIN_URL,
)
from pysmarthashtag.models import SmartAPIError

EXPIRES_AT_OFFSET = datetime.timedelta(seconds=HTTPX_TIMEOUT * 2)

_LOGGER = logging.getLogger(__name__)


class SmartAuthentication(httpx.Auth):
    """Authentication and Retry Handler for the Smart API."""

    def __init__(
        self,
        username: str,
        password: str,
        access_token: Optional[str] = None,
        expires_at: Optional[datetime.datetime] = None,
        refresh_token: Optional[str] = None,
    ):
        self.username: str = username
        self.password: str = password
        self.access_token: Optional[str] = access_token
        self.expires_at: Optional[datetime.datetime] = expires_at
        self.refresh_token: Optional[str] = refresh_token
        self.device_id: str = secrets.token_hex(8)
        self._lock: Optional[asyncio.Lock] = None
        self.api_access_token: Optional[str] = None
        self.api_refresh_token: Optional[str] = None
        self.api_user_id: Optional[str] = None
        _LOGGER.debug("Device ID: %s", self.device_id)

    @property
    def login_lock(self) -> asyncio.Lock:
        """Make sure there is only one login at a time."""
        if not self._lock:
            self._lock = asyncio.Lock()
        return self._lock

    def sync_auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        """Handle synchronous authentication flow for requests."""
        raise RuntimeError("Cannot use an async authentication class with httpx.Client")

    async def async_auth_flow(self, request: Request) -> AsyncGenerator[Request, Response]:
        """Asynchronous authentication flow for handling requests."""
        _LOGGER.debug("Handling request %s", request.url)
        # Get an initial login on first call
        async with self.login_lock:
            if not self.access_token:
                await self.login()
        request.headers["Authorization"] = f"Bearer {self.access_token}"

        response: httpx.Response = yield request

        if response.is_success:
            return

        await response.aread()

        retry_count = 0
        while (
            response.status_code == 429 or (response.status_code == 403 and "quota" in response.text.lower())
        ) and retry_count < 3:
            wait_time = get_retry_wait_time(response)
            _LOGGER.debug("Rate limit exceeded. Waiting %s seconds", wait_time)
            await asyncio.sleep(wait_time)
            response = yield request
            await response.aread()
            retry_count += 1

        if response.status_code == 401:
            async with self.login_lock:
                _LOGGER.debug("Token expired. Refreshing token")
                await self.login()
                request.headers["Authorization"] = f"Bearer {self.access_token}"

            _LOGGER.debug("Token expired. Refreshing token")
            await self.login()
            request.headers["Authorization"] = f"Bearer {self.access_token}"
            response = yield request
            await response.aread()

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            _LOGGER.error(
                "Error handling request %s: %s",
                request.url,
                exc,
            )
            raise

    async def login(self) -> None:
        """Login to the Smart API."""
        _LOGGER.debug("Logging in to Smart API")
        token_data = {}
        if self.refresh_token:
            token_data = await self._refresh_access_token()
        if not token_data:
            token_data = await self._login()
        try:
            token_data["expires_at"] = token_data["expires_at"] - EXPIRES_AT_OFFSET

            self.access_token = token_data["access_token"]
            self.refresh_token = token_data["refresh_token"]
            self.api_access_token = token_data["api_access_token"]
            self.api_refresh_token = token_data["api_refresh_token"]
            self.api_user_id = token_data["api_user_id"]
            self.expires_at = token_data["expires_at"]
            _LOGGER.debug(f"Login successful: {token_data}")
            return True
        except KeyError:
            raise SmartAPIError("Could not login to Smart API")

    async def _refresh_access_token(self):
        """Refresh the access token."""
        try:
            async with SmartLoginClient() as _:
                _LOGGER.debug("Refreshing access token via relogin because refresh token is not implemented")
                await self._login()
        except SmartAPIError:
            _LOGGER.debug("Refreshing access token failed. Logging in again")
            return {}

    async def _login(self):
        """Login to Smart web services."""
        async with SmartLoginClient() as client:
            _LOGGER.info("Aquiring access token.")

            # Get Context
            r_context = await client.get(
                CONTEXT_URL,
                headers={
                    "x-app-id": "SmartAPPEU",
                    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
                    "accept-language": "de-DE,de;q=0.9,en-DE;q=0.8,en-US;q=0.7,en;q=0.6",
                    "x-requested-with": "com.smart.hellosmart",
                    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
                    "content-type": "application/json; charset=utf-8",
                },
                follow_redirects=True,
            )
            try:
                context = r_context.url.params["context"]
                _LOGGER.debug("Context: %s", context)
            except KeyError:
                raise SmartAPIError("Could not get context from login page")

            # Get login token from Smart API
            r_login = await client.post(
                LOGIN_URL,
                data={
                    "loginID": self.username,
                    "password": self.password,
                    "sessionExpiration": 2592000,
                    "targetEnv": "jssdk",
                    "include": "profile%2Cdata%2Cemails%2Csubscriptions%2Cpreferences%2C",
                    "includeUserInfo": True,
                    "loginMode": "standard",
                    "lang": "de",
                    "APIKey": API_KEY,
                    "source": "showScreenSet",
                    "sdk": "js_latest",
                    "pageURL": "https%3A%2F%2Fapp.id.smart.com%2Flogin%3Fgig_ui_locales%3Dde-DE",
                    "sdkBuild": 15482,
                    "format": "json",
                    "riskContext": "%7B%22b0%22%3A41187%2C%22b1%22%3A%5B0%2C2%2C3%2C1%5D%2C%22b2%22%3A4%2C%22b3%22%3A%5B%22-23%7C0.383%22%2C%22-81.33333587646484%7C0.236%22%5D%2C%22b4%22%3A3%2C%22b5%22%3A1%2C%22b6%22%3A%22Mozilla%2F5.0%20%28Linux%3B%20Android%209%3B%20ANE-LX1%20Build%2FHUAWEIANE-L21%3B%20wv%29%20AppleWebKit%2F537.36%20%28KHTML%2C%20like%20Gecko%29%20Version%2F4.0%20Chrome%2F118.0.0.0%20Mobile%20Safari%2F537.36%22%2C%22b7%22%3A%5B%5D%2C%22b8%22%3A%2216%3A33%3A26%22%2C%22b9%22%3A-60%2C%22b10%22%3Anull%2C%22b11%22%3Afalse%2C%22b12%22%3A%7B%22charging%22%3Atrue%2C%22chargingTime%22%3Anull%2C%22dischargingTime%22%3Anull%2C%22level%22%3A0.58%7D%2C%22b13%22%3A%5B5%2C%22360%7C760%7C24%22%2Cfalse%2Ctrue%5D%7D",
                },
                headers={
                    "accept": "*/*",
                    "accept-language": "de",
                    "content-type": "application/x-www-form-urlencoded",
                    "x-requested-with": "com.smart.hellosmart",
                    "cookie": "gmid=gmid.ver4.AcbHPqUK5Q.xOaWPhRTb7gy-6-GUW6cxQVf_t7LhbmeabBNXqqqsT6dpLJLOWCGWZM07EkmfM4j.u2AMsCQ9ZsKc6ugOIoVwCgryB2KJNCnbBrlY6pq0W2Ww7sxSkUa9_WTPBIwAufhCQYkb7gA2eUbb6EIZjrl5mQ.sc3; ucid=hPzasmkDyTeHN0DinLRGvw; hasGmid=ver4; gig_bootstrap_3_L94eyQ-wvJhWm7Afp1oBhfTGXZArUfSHHW9p9Pncg513hZELXsxCfMWHrF8f5P5a=auth_ver4",
                    "origin": "https://app.id.smart.com",
                    "user-agent": "Hello smart/1.4.0 (iPhone; iOS 17.1; Scale/3.00)",
                },
            )
            try:
                login_result = r_login.json()
                login_token = login_result["sessionInfo"]["login_token"]
                expires_at = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
                    seconds=int(login_result["sessionInfo"]["expires_in"])
                )
            except (KeyError, ValueError):
                raise SmartAPIError("Could not get login token from login page")

            auth_url = AUTH_URL + "?context=" + context + "&login_token=" + login_token
            cookie = f"gmid=gmid.ver4.AcbHPqUK5Q.xOaWPhRTb7gy-6-GUW6cxQVf_t7LhbmeabBNXqqqsT6dpLJLOWCGWZM07EkmfM4j.u2AMsCQ9ZsKc6ugOIoVwCgryB2KJNCnbBrlY6pq0W2Ww7sxSkUa9_WTPBIwAufhCQYkb7gA2eUbb6EIZjrl5mQ.sc3; ucid=hPzasmkDyTeHN0DinLRGvw; hasGmid=ver4; gig_bootstrap_3_L94eyQ-wvJhWm7Afp1oBhfTGXZArUfSHHW9p9Pncg513hZELXsxCfMWHrF8f5P5a=auth_ver4; glt_3_L94eyQ-wvJhWm7Afp1oBhfTGXZArUfSHHW9p9Pncg513hZELXsxCfMWHrF8f5P5a={login_token}"
            r_auth = await client.get(
                auth_url,
                headers={
                    "accept": "*/*",
                    "cookie": cookie,
                    "accept-language": "de-DE,de;q=0.9,en-DE;q=0.8,en-US;q=0.7,en;q=0.6",
                    "x-requested-with": "com.smart.hellosmart",
                    "user-agent": "Mozilla/5.0 (Linux; Android 9; ANE-LX1 Build/HUAWEIANE-L21; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/118.0.0.0 Mobile Safari/537.36",
                },
            )
            if "location" not in r_auth.headers:
                raise SmartAPIError("Could not get location from auth page")

            r_auth = await client.get(
                r_auth.headers["location"],
                headers={
                    "accept": "*/*",
                    "cookie": cookie,
                    "accept-language": "de-DE,de;q=0.9,en-DE;q=0.8,en-US;q=0.7,en;q=0.6",
                    "x-requested-with": "com.smart.hellosmart",
                    "user-agent": "Mozilla/5.0 (Linux; Android 9; ANE-LX1 Build/HUAWEIANE-L21; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/118.0.0.0 Mobile Safari/537.36",
                },
            )
            try:
                auth_result = httpx.URL(r_auth.headers["location"])
                access_token = auth_result.params["access_token"]
                refresh_token = auth_result.params["refresh_token"]
            except KeyError:
                raise SmartAPIError("Could not get access token from auth page")

            data = json.dumps({"accessToken": access_token}).replace(" ", "")
            r_api_access = await client.post(
                API_BASE_URL + API_SESION_URL + "?identity_type=smart",
                headers={
                    **utils.generate_default_header(
                        self.device_id,
                        None,
                        params={
                            "identity_type": "smart",
                        },
                        method="POST",
                        url=API_SESION_URL,
                        body=data,
                    )
                },
                data=data,
            )
            api_result = r_api_access.json()
            _LOGGER.debug("API access result: %s", api_result)
            try:
                api_access_token = api_result["data"]["accessToken"]
                api_refresh_token = api_result["data"]["refreshToken"]
                api_user_id = api_result["data"]["userId"]
            except KeyError:
                raise SmartAPIError("Could not get API access token from API")

        return {
            "access_token": access_token,
            "refresh_token": refresh_token,
            "api_access_token": api_access_token,
            "api_refresh_token": api_refresh_token,
            "api_user_id": api_user_id,
            "expires_at": expires_at,
        }


class SmartLoginClient(httpx.AsyncClient):
    """Client to login to the Smart API."""

    def __init__(self, *args, **kwargs):
        # Increase timeout to 30 seconds
        kwargs["timeout"] = httpx.Timeout(HTTPX_TIMEOUT)

        # Register event hooks
        kwargs["event_hooks"] = defaultdict(list, **kwargs.get("event_hooks", {}))

        # Event hook for raise_for_status on all requests
        async def raise_for_status_handler(response: httpx.Response):
            """Eventhandler that automaticvalle raises HTTPStatusError when attached to a request.

            Only raise on 4xx/5xx errors but not on 429.
            """
            if response.is_error and response.status_code != 429:
                try:
                    response.raise_for_status()
                except httpx.HTTPStatusError as exc:
                    _LOGGER.error(
                        "Error handling request %s: %s",
                        response.url,
                        exc,
                    )
                    raise

        kwargs["event_hooks"]["response"].append(raise_for_status_handler)

        async def log_request(request):
            if request.method == "POST":
                await request.aread()
                _LOGGER.debug(
                    f"Request: {request.method} {request.url} - {request.content.decode()} - {request.headers}"
                )
            else:
                _LOGGER.debug(f"Request: {request.method} {request.url}")

        async def log_response(response):
            await response.aread()
            request = response.request
            _LOGGER.debug(f"Response: {request.method} {request.url} - Status {response.status_code}")

        kwargs["event_hooks"]["response"].append(log_response)
        kwargs["event_hooks"]["request"].append(log_request)

        super().__init__(**kwargs)


class SmartLoginRetry(httpx.Auth):
    """httpx.Auth uses as waorkauround to retry and sleep in 429."""

    def sync_auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        raise RuntimeError("Cannot use a async authentication class with httpx.Client")

    async def async_auth_flow(self, request: Request) -> AsyncGenerator[Request, Response]:
        # Try getting a response
        response: httpx.Response = yield request

        for _ in range(3):
            if response.status_code == 429:
                await response.aread()
                wait_time = get_retry_wait_time(response)
                _LOGGER.debug("Rate limit exceeded. Waiting %s seconds", wait_time)
                await asyncio.sleep(wait_time)
                response = yield request

                # Only checking for 429 errors, all other errors are raised in SmartLoginClient
                if response.status_code == 429:
                    try:
                        response.raise_for_status()
                    except httpx.HTTPStatusError as exc:
                        _LOGGER.error(
                            "Error handling request %s: %s",
                            request.url,
                            exc,
                        )
                        raise


def get_retry_wait_time(response: httpx.Response) -> int:
    """Get the wait time to wait twice as long before retrying."""
    try:
        retry_after = next(iter([int(i) for i in response.json().get("message", "") if i.isdigit()]))
    except Exception:
        retry_after = 2
    return math.ceil(retry_after * 2)



"""Fixtures for Smart tests."""

from typing import Generator

import pytest
import respx

from pysmarthashtag.account import SmartAccount

from . import (
    TEST_PASSWORD,
    TEST_USERNAME,
)
from .common import SmartMockRouter


@pytest.fixture
def smart_fixture(request: pytest.FixtureRequest) -> Generator[respx.MockRouter, None, None]:
    """Patch Smart login API calls."""
    # Now we can start patching the API calls
    router = SmartMockRouter()

    with router:
        yield router


async def prepare_account_with_vehicles():
    """Initialize account and get vehicles."""
    account = SmartAccount(TEST_USERNAME, TEST_PASSWORD)
    await account.get_vehicles()
    return account



"""URLs for different services and error code mapping."""

API_KEY = "3_L94eyQ-wvJhWm7Afp1oBhfTGXZArUfSHHW9p9Pncg513hZELXsxCfMWHrF8f5P5a"
SERVER_URL = "https://awsapi.future.smart.com/login-app/api/v1/authorize?uiLocales=de-DE"
CONTEXT_URL = "https://awsapi.future.smart.com/login-app/api/v1/authorize?uiLocales=de-DE&uiLocales=de-DE"
AUTH_URL = f"https://auth.smart.com/oidc/op/v1.0/{API_KEY}/authorize/continue"
LOGIN_URL = "https://auth.smart.com/accounts.login"
API_BASE_URL = "https://api.ecloudeu.com"
API_CARS_URL = "/device-platform/user/vehicle/secure"
API_SESION_URL = "/auth/account/session/secure"
API_SELECT_CAR_URL = "/device-platform/user/session/update"
API_TELEMATICS_URL = "/remote-control/vehicle/telematics/"

HTTPX_TIMEOUT = 30.0



"""Python package description."""
from setuptools import setup

setup(
    setup_requires=["pbr"],
    pbr=True,
    install_requires=[
        "httpx",
        "pycryptodome>=3.4",
        "pyjwt>=2.1.0",
    ],
)



"""Provides an accessible controll of the vehicle's climate functions."""

import json
import logging
from enum import Enum
from typing import TypedDict

from pysmarthashtag.api import utils
from pysmarthashtag.api.client import SmartClient
from pysmarthashtag.const import API_BASE_URL, API_TELEMATICS_URL
from pysmarthashtag.models import SmartHumanCarConnectionError, SmartTokenRefreshNecessary

_LOGGER = logging.getLogger(__name__)


class ServiceParameter(TypedDict):
    """TypedDict for service parameters."""

    key: str
    value: str


class HeatingLocation(str, Enum):
    """Enum for heating locations in the vehicle."""

    DRIVER_SEAT = "front-left"
    PASSENGER_SEAT = "front-right"
    STEERING_WHEEL = "steering_wheel"


class ClimateControll:
    """Provides an accessible controll of the vehicle's climate functions."""

    BASE_PAYLOAD_TEMPLATE = {
        "creator": "tc",
        "operationScheduling": {
            "duration": 180,
            "interval": 0,
            "occurs": 1,
            "recurrentOperation": False,
        },
        "serviceId": "RCE_2",
    }

    def __init__(self, config, vin):
        """Initialize the vehicle."""
        self.config = config
        self.vin = vin
        self.conditioning_temp = 20.0
        self.heating_levels: dict[HeatingLocation, int] = {
            HeatingLocation.DRIVER_SEAT: 0,
            HeatingLocation.PASSENGER_SEAT: 0,
            HeatingLocation.STEERING_WHEEL: 0,
        }

    def _get_payload(self, active: bool) -> str:
        _payload = self.BASE_PAYLOAD_TEMPLATE.copy()
        _payload["command"] = "stop" if not active else "start"
        _payload["timestamp"] = utils.create_correct_timestamp()
        _payload["serviceParameters"] = []
        _payload["serviceParameters"].append({"key": "rce.conditioner", "value": "1"})
        _payload["serviceParameters"].append({"key": "rce.temp", "value": f"{self.conditioning_temp:.1f}"})
        for loc, level in self.heating_levels.items():
            if level > 0:
                _payload["serviceParameters"] += self._add_rce_heating_service(loc, f"{level}")

        return json.dumps(_payload).replace(" ", "")

    def _add_rce_heating_service(self, value: str, level: int) -> list[ServiceParameter]:
        """Create heating service parameters for the payload.

        Args:
        ----
            value: The heating location identifier
            level: The heating level as a string

        Returns:
        -------
            List of service parameter dictionaries

        """
        payload = [{"key": "rce.heat", "value": value}, {"key": "rce.level", "value": level}]
        return payload

    def set_heating_level(self, location: HeatingLocation, level: int):
        """Set heating level for the specified location."""
        if not isinstance(level, int):
            raise TypeError("Heating level must be an integer")
        if level > 3 or level < 0:
            raise ValueError("Seat heating level must be between 0 and 3.")
        self.heating_levels[location] = level

    async def set_climate_conditioning(self, temp: float, active: bool) -> bool:
        """Set the climate conditioning."""

        if not isinstance(temp, (int, float)):
            raise TypeError("Temperature must be a number")
        if temp < 16 or temp > 30:
            raise ValueError("Temperature must be between 16 and 30 degrees.")
        self.conditioning_temp = float(temp)

        async with SmartClient(self.config) as client:
            params = self._get_payload(active)
            _LOGGER.debug(f"Setting climate conditioning: {params}")
            for retry in range(2):
                try:
                    vehicles_response = await client.put(
                        API_BASE_URL + API_TELEMATICS_URL + self.vin,
                        headers={
                            **utils.generate_default_header(
                                client.config.authentication.device_id,
                                client.config.authentication.api_access_token,
                                params={},
                                method="PUT",
                                url=API_TELEMATICS_URL + self.vin,
                                body=params,
                            )
                        },
                        data=params,
                    )
                    api_result = vehicles_response.json()
                    return api_result["success"]
                except SmartTokenRefreshNecessary:
                    _LOGGER.debug(f"Got Token Error, retry: {retry}")
                    continue
                except SmartHumanCarConnectionError:
                    _LOGGER.debug(f"Got Human Car Connection Error, retry: {retry}")
                    continue
